from yapsy.IPlugin import IPlugin
from yapsy.PluginManager import PluginManager

import sys,os
import json
import codecs

def debug(msg):
    DEBUG=True
    if DEBUG:
        print(msg,file=sys.stderr)

def get_params(*args,**kwargs):
    paramnames=kwargs['paramnames']
    params={}
    for x in paramnames:
        try:
            idx=paramnames.index(x)
            params[x] = args[idx]
        except:
            try:
                params[x] = kwargs[x]
            except:
                raise Exception('ErrorParameter')
    return params

class IPlugin(IPlugin):
    def __init__(self):
        super().__init__()
        self.syncer=None
        self.is_blocker = False
        self.typeplug='Base'
        self.params=['param1','param2']
        self.data_input = {}
        self.data_output = {}
        self.init_plugin()

    def syntax(self):
        return [self.typeplug] + self.params

    def activate(self,*args,**kwargs):
        super().activate()

    def deactivate(self,*args,**kwargs):
        super().deactivate()

    def input(self,*args,**kwargs):
        if not self.is_activated:
            self.activate()
        dparam={}
        params = get_params(*args,paramnames=['trans_id','params'], **kwargs)
        iterparam=iter(params['params'])

        for param in self.params:
                dparam[param]=next(iterparam)
        self.store_input(trans_id=params['trans_id'],**dparam)

    def store_output(self,transid,status,output_list):
        if type(status) != type(bool()) or type(output_list) != type(list()):
            raise Exception('Error store_output on %s plugin!'%self.typeplug)
        self.data_output[transid]={'status':status,'output':output_list}

    def store_input(self,*args,**kwargs):
        dparam={}
        params = get_params(*args,  paramnames=['trans_id'],**kwargs)
        for paramname in self.params:
            dparam[paramname]=kwargs[paramname]
        self.data_input[params['trans_id']]=dparam


    def end_process(self,*args,**kwargs):
        params = get_params(*args, paramnames=['transid','status', 'output'], **kwargs)
        self.store_output(params['transid'], params['status'], [params['output']])


    def init_process(self,*args,**kwargs):
        try:
            transid=kwargs['transid']
            status,output=self.process(*args,params=self.data_input[transid], **kwargs)
        except Exception as e:
            status=False
            output=str(e)
        self.end_process(*args,status=status,output=output, **kwargs)

    def process(self,*args,**kwargs):
        debug('Using process (transid {}): on {} '.format(kwargs['transid'],self.typeplug ))
        return True,[ 'Dummy' ,kwargs['transid'] ]

    def output(self,*args,**kwargs):
        if not self.is_activated:
            raise Exception('Plugin %s not activated'%self.typeplug)
        params = get_params(*args, paramnames=['transid'], **kwargs)
        return self.data_output[params['transid']]

class SyncerClass(PluginManager):

    def __init__(self,plugin_path=['plugins'],ext='syncplugin'):
        super().__init__(categories_filter={'Default':IPlugin},plugin_info_ext=ext,directories_list=plugin_path)
        self.syncplugins={}
        self.sid=0

        self.collectPlugins()
        self.reorder_plugins()

    def get_plugin(self,pluginname):
        return self.syncplugins[pluginname]['obj']

    def get_parameters_from_plugin(self,pluginname):
        return self.syncplugins[pluginname]['params']

    def action_plugins(self,action=None,typeplug=None):
        if typeplug == None:
            for plugin in (self.syncplugins[plugintype] for plugintype in self.syncplugins):
                if action == 'activate':
                    plugin['obj'].activate()
                else:
                    plugin['obj'].deactivate()

    def deactivate_plugins(self,typeplug=None):
        self.action_plugins('deactivate')

    def activate_plugins(self,typeplug=None):
        self.action_plugins('activate')

    def reorder_plugins(self):
        for plugin in self.getAllPlugins():
     #       plugin.plugin_object.make_attributes(plugin.plugin_object.params)
            # TODO: realizar metodo en syncer para evitar el acceso al syncer , se podria enganchar una funcion que permita ejecutar algunas cosas en syncer despues de un filtrado, en vez de todo el syncer
            plugin.plugin_object.syncer=self
            self.syncplugins[plugin.plugin_object.typeplug]={'obj':plugin.plugin_object,'params':plugin.plugin_object.params,'trans_ids':[]}

    def get_info_parser(self):
        dparser_info={}
        for plugtype in self.syncplugins:
            dparser_info[plugtype]=self.syncplugins[plugtype]['params']
        return {'choices': list(self.syncplugins.keys()), 'parser_info': dparser_info}

    def get_active_plugins(self):
        lactive=[]
        for plugtype in self.syncplugins:
            if self.get_plugin(plugtype).is_activated:
                lactive.append(plugtype)
        return lactive
    def get_plugin_from_transid(self,transid):
        for pluginname in self.get_active_plugins():
            if transid in self.syncplugins[pluginname]['trans_ids']:
                return pluginname
        return None
    def get_pluginobj_from_transid(self,transid):
        #Equivalente de get_plugin_from_transid + get_plugin
        for pluginname in self.get_active_plugins():
            if transid in self.syncplugins[pluginname]['trans_ids']:
                return self.syncplugins[pluginname]['obj']
        return None
    def plugin_input(self,plugin,params):
        self.sid+=1
        self.syncplugins[plugin]['trans_ids'].append(self.sid)
        self.syncplugins[plugin]['obj'].input(self.sid,params)

class SyncerSession(object):
    def __init__(self,*args,**kwargs):
        params=get_params(*args,paramnames=['path','lock','file'], **kwargs)
        self.syncpath=params['path']
        self.synclockfile=params['lock']
        self.syncsessionfile=params['file']

    def destroy(self,*args,**kwargs):
        self.validate_session(msg='Session not initiated')
        os.remove(self.synclockfile)
        os.remove(self.syncsessionfile)

    def init(self,*args,**kwargs):
        if not os.path.isdir(self.syncpath):
            os.mkdir(self.syncpath,mode=0o777)
        else:
            if os.path.isfile(self.synclockfile):
                raise Exception('Session already initiated')
            else:
                open(self.synclockfile, 'w').close()
                with open(self.syncsessionfile, 'w') as f:
                    json.dump([], f, separators=(',', ':'))

    def put_in_session(self,*args, **kwargs):
        self.validate_session(msg='Error writting into session')
        params = get_params(*args, paramnames=['data'], **kwargs)
        with open(self.syncsessionfile, 'r+') as f:
            jsdata = json.load(f)
            lresult = []
            plugin = params['data']['plugin']
            params = params['data']['content']
            lresult.append((plugin, params))
            jsdata.extend(lresult)
            f.seek(0)
            json.dump(jsdata, f, separators=(',', ':'))

    def put_file_in_session(self,*args, **kwargs):
        self.validate_session()
        params = get_params(*args,paramnames=['data_file'], **kwargs)

        if type(params['data_file']) == type(str()):
            f = open(params['data_file'], 'rb')
        else:
            f = params['data_file']

        reader = codecs.getreader('utf-8')(f)
        jsdata = json.load(reader)
        f.close()

        with open(self.syncsessionfile, 'r') as sf:
            jsdata_session = json.load(sf)
            jsdata_session.extend(jsdata)
        with open(self.syncsessionfile, 'w') as sf:
            json.dump(jsdata_session, sf, separators=(',', ':'))

    def validate_session(self,*args, **kwargs):
        if not os.path.isdir(self.syncpath) or not os.path.isfile(self.syncsessionfile) or not os.path.isfile(self.synclockfile):
            if 'msg' in kwargs and type(kwargs['msg']) == type(str()):
                raise Exception(kwargs['msg'])
            else:
                raise Exception('Error reading session')
        return True

    def get_session_data(self,*args, **kwargs):
        self.validate_session()
        with open(self.syncsessionfile, 'r') as f:
            return f.readlines()

    def load_session(self,*args, **kwargs):
        self.validate_session()
        with open(self.syncsessionfile, 'r') as f:
            jsdata = json.load(f)
        return jsdata
